import logging
from typing import Optional

from monoid_pydev.models.models import MonoidLogMessage, MonoidMessage


def get_logger(name: Optional[str] = None):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(MonoidFormatter())

    logger.addHandler(stream_handler)

    return logger


class MonoidFormatter(logging.Formatter):
    """
    Converts log messages into MonoidMessage json objects.
    """

    # Transforming Python log levels to Airbyte protocol log levels
    level_mapping = {
        logging.FATAL: "FATAL",
        logging.ERROR: "ERROR",
        logging.WARNING: "WARN",
        logging.INFO: "INFO",
        logging.DEBUG: "DEBUG",
    }

    def format(self, record) -> str:
        level = self.level_mapping.get(record.levelno, "INFO")
        message = super().format(record)
        return MonoidMessage(
            type="LOG",
            log=MonoidLogMessage(
                message=message
            )
        ).json()



from typing import Dict, Any, Iterable, Optional

from monoid_pydev.models import (
    MonoidRecord, MonoidSchema, MonoidQueryIdentifier, RequestStatus, DataType,
    MonoidRequestHandle, MonoidRequestResult
)
from abc import ABC, abstractmethod

from monoid_pydev.models.models import MonoidPersistenceConfig, MonoidRequestStatus, RequestType
from monoid_pydev.silos.data_store import DataStore


class DBDataStore(DataStore):
    @abstractmethod
    def query_records(
        self,
        persistence_conf: MonoidPersistenceConfig,
        query: MonoidQueryIdentifier,
    ) -> Iterable[MonoidRecord]:
        """
        To be implemented by subclasses.
        """

    @abstractmethod
    def delete_records(
        self,
        persistence_conf: MonoidPersistenceConfig,
        query: MonoidQueryIdentifier,
    ):
        """
        To be implemented by subclasses.
        """

    @abstractmethod
    def scan_records(
        self,
        persistence_conf: MonoidPersistenceConfig,
        schema: MonoidSchema,
    ):
        """
        To be implemented by subclasses.
        """

    def run_query_request(
        self,
        persistence_conf: MonoidPersistenceConfig,
        query: MonoidQueryIdentifier,
    ) -> MonoidRequestResult:
        """
        The base behavior for a DB Data store is to return that the request
        is complete, and run the query in scan_records.
        """

        return MonoidRequestResult(
            status=MonoidRequestStatus(
                schema_group=self.group(),
                schema_name=self.name(),
                request_status=RequestStatus.COMPLETE,
                data_type=DataType.RECORDS
            ),
            handle=MonoidRequestHandle(
                schema_group=self.group(),
                schema_name=self.name(),
                request_type=RequestType.QUERY,
                data={
                    "query": query
                }
            )
        )

    def run_delete_request(
        self,
        persistence_conf: MonoidPersistenceConfig,
        query: MonoidQueryIdentifier,
    ) -> MonoidRequestResult:
        """
        Starts a delete request
        """

        self.delete_records(persistence_conf, query)
        
        return MonoidRequestResult(
            status=MonoidRequestStatus(
                schema_group=self.group(),
                schema_name=self.name(),
                request_status=RequestStatus.COMPLETE,
                data_type=DataType.NONE
            ),
            handle=MonoidRequestHandle(
                schema_group=self.group(),
                schema_name=self.name(),
                request_type=RequestType.DELETE,
                data={
                    "query": query
                }
            )
        )

    def request_status(
        self,
        persistence_conf: MonoidPersistenceConfig,
        handle: MonoidRequestHandle
    ) -> MonoidRequestStatus:
        """
        Gets the status of a request
        """

        if handle.request_type == RequestType.QUERY:
            return MonoidRequestStatus(
                schema_group=self.group(),
                schema_name=self.name(),
                request_status=RequestStatus.COMPLETE,
                data_type=DataType.RECORDS
            )
        elif handle.request_type == RequestType.DELETE:
            return MonoidRequestStatus(
                schema_group=self.group(),
                schema_name=self.name(),
                request_status=RequestStatus.COMPLETE,
                data_type=DataType.NONE
            )

        raise ValueError(f"Unknown request type {handle.request_type}")

    def request_results(
        self,
        persistence_conf: MonoidPersistenceConfig,
        handle: MonoidRequestHandle
    ) -> Iterable[MonoidRecord]:
        """
        Gets the result of a request
        """

        if handle.data is None:
            return

        if handle.request_type == RequestType.QUERY:
            query = MonoidQueryIdentifier.parse_obj(handle.data["query"])
            yield from self.query_records(persistence_conf, query)
            return
        elif handle.request_type == RequestType.DELETE:
            return

        raise ValueError(f"Unknown request type {handle.request_type}")



# generated by datamodel-codegen:
#   filename:  monoid_protocol.json
#   timestamp: 2022-11-28T06:45:50+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel


class MonoidQueryIdentifier(BaseModel):
    schema_name: str
    schema_group: Optional[str] = None
    identifier: str
    identifier_query: Union[str, int]
    json_schema: Dict[str, Any]


class RecordType(Enum):
    RECORD = 'RECORD'
    FILE = 'FILE'


class MonoidRecord(BaseModel):
    record_type: Optional[RecordType] = None
    schema_name: str
    schema_group: Optional[str] = None
    data: Optional[Dict[str, Any]] = None
    file: Optional[str] = None


class MonoidSchema(BaseModel):
    name: str
    group: Optional[str] = None
    json_schema: Dict[str, Any]


class MonoidSiloSpec(BaseModel):
    name: Optional[str] = None
    spec: Dict[str, Any]


class MonoidSchemasMessage(BaseModel):
    schemas: List[MonoidSchema]


class Status(Enum):
    SUCCESS = 'SUCCESS'
    FAILURE = 'FAILURE'


class MonoidValidateMessage(BaseModel):
    status: Status
    message: Optional[str] = None


class MonoidLogMessage(BaseModel):
    message: str


class MonoidPersistenceConfig(BaseModel):
    temp_store: str


class RequestStatus(Enum):
    PROGRESS = 'PROGRESS'
    COMPLETE = 'COMPLETE'
    FAILED = 'FAILED'


class DataType(Enum):
    RECORDS = 'RECORDS'
    FILE = 'FILE'
    NONE = 'NONE'


class MonoidRequestStatus(BaseModel):
    schema_group: Optional[str] = None
    schema_name: str
    request_status: RequestStatus
    data_type: Optional[DataType] = None


class RequestType(Enum):
    QUERY = 'QUERY'
    DELETE = 'DELETE'


class MonoidRequestHandle(BaseModel):
    schema_group: Optional[str] = None
    schema_name: str
    data: Optional[Dict[str, Any]] = None
    request_type: RequestType


class MonoidRequestsMessage(BaseModel):
    handles: List[MonoidRequestHandle]


class Type(Enum):
    SCHEMA = 'SCHEMA'
    RECORD = 'RECORD'
    SPEC = 'SPEC'
    REQUEST_RESULT = 'REQUEST_RESULT'
    REQUEST_STATUS = 'REQUEST_STATUS'
    VALIDATE = 'VALIDATE'
    LOG = 'LOG'


class MonoidQuery(BaseModel):
    identifiers: List[MonoidQueryIdentifier]


class MonoidRequestResult(BaseModel):
    status: MonoidRequestStatus
    handle: MonoidRequestHandle


class MonoidMessage(BaseModel):
    type: Type
    record: Optional[MonoidRecord] = None
    schema_msg: Optional[MonoidSchemasMessage] = None
    spec: Optional[MonoidSiloSpec] = None
    validate_msg: Optional[MonoidValidateMessage] = None
    log: Optional[MonoidLogMessage] = None
    request: Optional[MonoidRequestResult] = None
    request_status: Optional[MonoidRequestStatus] = None


class MonoidProtocol(BaseModel):
    MonoidMessage: Optional[MonoidMessage] = None



from typing import Iterable, List

import jsonschema
from monoid_pydev.models.models import MonoidPersistenceConfig
from monoid_pydev.silos import AbstractSilo
import argparse
from jsonschema import validate
from monoid_pydev.models import MonoidMessage, Type


class MonoidRunner():
    def __init__(self, silo: AbstractSilo):
        self.silo = silo

    def parse_args(self, args: List[str]):
        parser = argparse.ArgumentParser()

        authed_parser = argparse.ArgumentParser(add_help=False)
        authed_parser.add_argument("-c", "--config", required=True)

        persistence_parser = argparse.ArgumentParser(add_help=False)
        persistence_parser.add_argument("-p", "--persist_conf", required=True)

        subparsers = parser.add_subparsers(dest='command', required=True)

        scan_parser = subparsers.add_parser(
            "scan", parents=[authed_parser, persistence_parser])
        scan_parser.add_argument(
            "-s", "--schemas", required=True
        )

        delete_parser = subparsers.add_parser(
            "delete", parents=[authed_parser, persistence_parser])
        delete_parser.add_argument(
            "-q", "--query", required=True)

        query_parser = subparsers.add_parser(
            "query",
            parents=[authed_parser, persistence_parser]
        )
        query_parser.add_argument(
            "-q", "--query", required=True)

        req_results_parser = subparsers.add_parser(
            "request-results",
            parents=[authed_parser, persistence_parser]
        )
        req_results_parser.add_argument(
            "-r", "--requests", required=True)

        req_status_parser = subparsers.add_parser(
            "request-status",
            parents=[authed_parser, persistence_parser]
        )
        req_status_parser.add_argument(
            "-r", "--requests", required=True)

        subparsers.add_parser("schema", parents=[authed_parser])
        subparsers.add_parser("validate", parents=[authed_parser])
        subparsers.add_parser("spec")

        self.parse_result = parser.parse_args(args)

    def run(self) -> Iterable[str]:
        spec = self.silo.spec()
        if self.parse_result.command == "spec":
            yield MonoidMessage(type=Type.SPEC, spec=spec).json()
            return

        config = self.silo.parse_config(self.parse_result.config)

        try:
            validate(config, spec.spec)
        except jsonschema.exceptions.ValidationError:
            raise ValueError("Invalid config")

        if self.parse_result.command == "schema":
            yield MonoidMessage(type=Type.SCHEMA, schema_msg=self.silo.schemas(config)).json()
            return

        elif self.parse_result.command == "validate":
            yield MonoidMessage(type=Type.VALIDATE, validate_msg=self.silo.validate(config)).json()
            return

        persist_conf = MonoidPersistenceConfig.parse_file(
            self.parse_result.persist_conf)

        if self.parse_result.command == "scan":
            schemas = self.silo.parse_schema(self.parse_result.schemas)
            for s in self.silo.scan(config, persist_conf, schemas):
                yield MonoidMessage(type=Type.RECORD, record=s).json()

        elif self.parse_result.command == "delete":
            query = self.silo.parse_query(self.parse_result.query)

            for req in self.silo.delete(config, persist_conf, query):
                yield MonoidMessage(type=Type.REQUEST_RESULT, request=req).json()

        elif self.parse_result.command == "query":
            query = self.silo.parse_query(self.parse_result.query)

            for req in self.silo.query(config, persist_conf, query):
                yield MonoidMessage(type=Type.REQUEST_RESULT, request=req).json()

        elif self.parse_result.command == "request-results":
            requests = self.silo.parse_requests(self.parse_result.requests)

            for rec in self.silo.request_results(config, persist_conf, requests):
                yield MonoidMessage(type=Type.RECORD, record=rec).json()

        elif self.parse_result.command == "request-status":
            requests = self.silo.parse_requests(self.parse_result.requests)

            for status in self.silo.request_status(config, persist_conf, requests):
                yield MonoidMessage(type=Type.REQUEST_STATUS, request_status=status).json()


def run_query(silo: AbstractSilo, args: List[str]):
    runner = MonoidRunner(silo)
    runner.parse_args(args)
    for res in runner.run():
        print(res)



from typing import Dict, Any, Iterable, Optional

from monoid_pydev.models import MonoidRecord, MonoidSchema, MonoidQueryIdentifier
from abc import ABC, abstractmethod

from monoid_pydev.models.models import MonoidPersistenceConfig, MonoidRequestHandle, MonoidRequestResult, MonoidRequestStatus


class DataStore(ABC):
    def to_brist_schema(self):
        return MonoidSchema(
            name=self.name(),
            group=self.group(),
            json_schema=self.json_schema()
        )

    @abstractmethod
    def name(self) -> str:
        """
        Get the name of the datastore.
        """

    @abstractmethod
    def group(self) -> Optional[str]:
        """
        Get the group of the datastore.
        """

    @abstractmethod
    def json_schema(self) -> Dict[str, Any]:
        """
        Returns the JSON schema of the data store.
        """

    @abstractmethod
    def run_query_request(
        self,
        persistence_conf: MonoidPersistenceConfig,
        query: MonoidQueryIdentifier,
    ) -> MonoidRequestResult:
        """
        Starts a query request
        """

    @abstractmethod
    def run_delete_request(
        self,
        persistence_conf: MonoidPersistenceConfig,
        query: MonoidQueryIdentifier,
    ) -> MonoidRequestResult:
        """
        Starts a delete request
        """

    @abstractmethod
    def request_status(
        self,
        persistence_conf: MonoidPersistenceConfig,
        handle: MonoidRequestHandle
    ) -> MonoidRequestStatus:
        """
        Gets the status of a request
        """

    @abstractmethod
    def request_results(
        self,
        persistence_conf: MonoidPersistenceConfig,
        handle: MonoidRequestHandle
    ) -> Iterable[MonoidRecord]:
        """
        Gets the result of a request
        """

    @abstractmethod
    def scan_records(
        self,
        persistence_conf: MonoidPersistenceConfig,
        schema: MonoidSchema,
    ):
        """
        To be implemented by subclasses.
        """



from abc import ABC, abstractmethod
import json
from re import S
from typing import Any, Iterable, Mapping, List, Optional
from monoid_pydev.models.models import MonoidRequestResult, MonoidRequestStatus, MonoidRequestsMessage

from monoid_pydev.silos.data_store import DataStore
from monoid_pydev.models import (
    MonoidQuery, MonoidRecord, MonoidSchema,
    MonoidSiloSpec, MonoidSchemasMessage, MonoidValidateMessage,
    MonoidPersistenceConfig, MonoidRequestHandle
)
import monoid_pydev.utils as utils


class AbstractSilo(ABC):
    @abstractmethod
    def data_stores(
        self,
        conf: Mapping[str, Any],
    ) -> List[DataStore]:
        """
        Gets the data stores that make up this silo.
        """

    def _data_stores_map(self, conf: Mapping[str, Any]):
        return {(d.group(), d.name()): d for d in self.data_stores(conf=conf)}

    def schemas(self, conf: Mapping[str, Any]) -> MonoidSchemasMessage:
        """
        Returns the set of schemas that are present in a data silo.
        """

        return MonoidSchemasMessage(schemas=[
            d.to_brist_schema() for d in self.data_stores(conf)
        ])

    def spec(self) -> MonoidSiloSpec:
        """
        Returns the spec for the silo.
        """
        f = utils.load_package_file(
            self.__class__.__module__.split(".")[0], "spec.json")
        if f is None:
            raise FileNotFoundError

        return MonoidSiloSpec.parse_obj(json.loads(f))

    def query(
        self,
        conf: Mapping[str, Any],
        persistence_conf: MonoidPersistenceConfig,
        query: MonoidQuery,
    ) -> Iterable[MonoidRequestResult]:
        """
        Queries records from a data silo based on a specific query.
        """
        data_stores = self._data_stores_map(conf)

        for query_rule in query.identifiers or []:
            data_store = data_stores[(
                query_rule.schema_group, query_rule.schema_name)]

            yield data_store.run_query_request(
                persistence_conf,
                query_rule
            )

    def delete(
        self,
        conf: Mapping[str, Any],
        persistence_conf: MonoidPersistenceConfig,
        query: MonoidQuery
    ) -> Iterable[MonoidRequestResult]:
        """
        Starts a monoid request that deletes records from the data silo
        based on a given query.
        """

        data_stores = self._data_stores_map(conf)

        for query_rule in query.identifiers:
            data_store = data_stores[(
                query_rule.schema_group, query_rule.schema_name)]

            yield data_store.run_delete_request(
                persistence_conf,
                query_rule
            )

    def request_results(
        self,
        conf: Mapping[str, Any],
        persistence_conf: MonoidPersistenceConfig,
        requests: MonoidRequestsMessage,
    ) -> Iterable[MonoidRecord]:
        data_stores = self._data_stores_map(conf)

        for handle in requests.handles:
            data_store = data_stores[(
                handle.schema_group, handle.schema_name)]

            yield from data_store.request_results(
                persistence_conf,
                handle
            )

    def request_status(
        self,
        conf: Mapping[str, Any],
        persistence_conf: MonoidPersistenceConfig,
        requests: MonoidRequestsMessage,
    ) -> Iterable[MonoidRequestStatus]:
        """
        Gets the status of a request.
        """

        data_stores = self._data_stores_map(conf)

        for handle in requests.handles:
            data_store = data_stores[(
                handle.schema_group, handle.schema_name)]

            yield data_store.request_status(
                persistence_conf,
                handle,
            )

    def scan(
        self,
        conf: Mapping[str, Any],
        persistence_conf: MonoidPersistenceConfig,
        schemas: MonoidSchemasMessage,
    ) -> Iterable[MonoidRecord]:
        """
        Returns a sample of the records in the data silo, to be used for data
        scanning.
        """

        data_stores = {
            (d.group(), d.name()): d for d in self.data_stores(
                conf=conf,
            )}

        for schema in schemas.schemas:
            data_store = data_stores[(
                schema.group, schema.name
            )]

            yield from data_store.scan_records(persistence_conf, schema)

    @abstractmethod
    def validate(
        self,
        conf: Mapping[str, Any],
    ) -> MonoidValidateMessage:
        """
        Validates that the configuration is set up correctly.
        """

    def parse_config(
        self,
        conf_file: str
    ) -> Mapping[str, Any]:
        """
        Parse a config file.
        """
        with open(conf_file, "r") as f:
            return json.loads(f.read())

    def parse_requests(
        self,
        requests_file: str
    ) -> MonoidRequestsMessage:
        """
        Parse a MonoidRequestsMessage file
        """
        return MonoidRequestsMessage.parse_file(requests_file)

    def parse_schema(
        self,
        schema_file: str
    ) -> MonoidSchemasMessage:
        """
        Parse a list of schemas.
        """
        return MonoidSchemasMessage.parse_file(schema_file)

    def parse_query(
        self,
        query_file: str
    ) -> MonoidQuery:
        """
        Parse a query.
        """
        return MonoidQuery.parse_file(query_file)
